#include "smptestconf.h"
/*
 * Copyright (C) by University of Illinois 2023
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "mpi.h"
#define BENV_WITHMPI 1
#include "benvutil.h"

static int verbose = 0, wrank=-1;
#define DBG0(str) DBGC0(str,wrank==0)
#define DBG1(str,a) DBGC1(str,a,wrank==0)
#define DBGC0(str,C) do { if (verbose && (C)) {	\
		       fprintf(stdout, str); fflush(stdout); } } while (0)
#define DBGC1(str,a,C) do { if (verbose && (C)) { \
			 fprintf(stdout, str, a); \
			 fflush(stdout); } } while (0)
#define DBGC2(str,a,b,C) do { if (verbose && (C)) {	\
			   fprintf(stdout, str, a, b); \
			   fflush(stdout); } } while (0)
#define DBGC3(str,a,b,cc,C) do { if (verbose && (C)) {		\
			      fprintf(stdout, str, a, b, cc); \
			      fflush(stdout); } } while (0)

static const char *hwdescline(int indent, int csize, int wrank, int idx,
			      const char *descstr);
static const char *hwdescrange(int indent, int range[3], const char *descstr);
static const char *hwdesclist(int indent, int n, int vec[],
			      const char *descstr);
static int compressListToRange(int n, int vec[], int range[3]);

#if MPI_VERSION < 4
int BENV_HwdescGetLocal(MPI_Comm comm, hwdesc_t hwdesc[], int maxdepth,
			    int *depth)
{
    return 1;
}
#else
/*@
  BENV_HwdescGetLocal - Using MPI, get information on the HW heirarchy

Input parameters:
+ comm - Communicator of processes
- maxdepth - Size of the 'hwdesc' array

Output Parameters:
+ hwdesc - Array describing the hardware heirarchy
- depth - Pointer to integer giving the depth of the hwdesc array (number of
 entries) I.e., hwdesc[0] ... hwdesc[depth-1] are valid

Notes:
While this routine returns only the local information for the calling rank,
it is a collective call over 'comm', and it includes information about
processes that are part of the same level of heirarchy, such as processes
with part of the same node.

This routine requires MPI Version of 4 or greater.  Returns failure (1)
if the MPI Version is less than 4.
  @*/
int BENV_HwdescGetLocal(MPI_Comm comm, hwdesc_t hwdesc[], int maxdepth,
			int *depth)
{
    int hwlevel = 0;

    hwdesc[0].comm    = comm;
    /* Try to get the communicator name as the description. */
    {
	char cname[MPI_MAX_OBJECT_NAME];
	int  cnameLen;
	MPI_Comm_get_name(comm, cname, &cnameLen);
	if (cnameLen > 0)
	    hwdesc[0].descstr = strdup(cname);
	else
	    hwdesc[0].descstr = strdup("Top level communicator");
    }

    /* Extract the hw_unguided comms */
    while (hwdesc[hwlevel].comm != MPI_COMM_NULL) {
	MPI_Info hwinfo;
	int      buflen, flag, lrank;
	char     *value;

	/* Get the number of distinct communicators at this level. This is
	   the number of processes that have rank 0 in a communicator at this
	   level.
	   Also gather the ranks in the parent communicator of the leaders
	   (processes with rank 0 in a communicator at this level)
	*/
	MPI_Comm_rank(hwdesc[hwlevel].comm, &lrank);
	if (hwlevel == 0) {
	    hwdesc[hwlevel].nDistinct  = 1;
	    hwdesc[hwlevel].idx        = 0; /* Only comm at this level */
	    hwdesc[hwlevel].leaders    = 0; /* No parent so no leaders array */
	    hwdesc[hwlevel].allsizeone = 0;
	}
	else {
	    BENVi_FindLeadersInSplit(hwdesc[hwlevel-1].comm,
				     hwdesc[hwlevel].comm, &hwdesc[hwlevel]);
#if 0
	    int *allleaders, *leaders, p_size, p_rank, is_lead;

	    DBG1("leaders for level %d\n", hwlevel);
	    is_lead = lrank == 0;
	    MPI_Allreduce(&is_lead, &hwdesc[hwlevel].nDistinct, 1, MPI_INT,
			  MPI_SUM, hwdesc[hwlevel-1].comm);
	    DBG1("Number of distinct comms is %d\n", hwdesc[hwlevel].nDistinct);

	    /* Gather up the rank in the parent of the leaders of the
	       communicators at this level. Algorithm is a simple
	       brute-force approach - gather ranks for *all* processes
	       at this level (array size of the parent comm) and look for
	       ranks >= 0 (use -1 if a process is not a leader on this
	       level) */
	    MPI_Comm_size(hwdesc[hwlevel-1].comm, &p_size);
	    allleaders = (int *)malloc(p_size * sizeof(int));
	    if (!allleaders) {
		MPI_Abort(MPI_COMM_WORLD, 1);
	    }
	    if (is_lead)
		MPI_Comm_rank(hwdesc[hwlevel-1].comm, &p_rank);
	    else
		p_rank = -1;
	    DBG0("Gather ranks of leaders in parent\n");
	    MPI_Gather(&p_rank, 1, MPI_INT, allleaders, 1, MPI_INT, 0,
		       hwdesc[hwlevel-1].comm);

	    /* On the root, compress this to remove the -1, then broadcast
	       to the rest of the processes */
	    DBG0("Bcast leaders\n");
	    leaders = (int *)malloc(hwdesc[hwlevel].nDistinct * sizeof(int));
	    if (!leaders) {
		MPI_Abort(MPI_COMM_WORLD, 0);
	    }
	    /* See above - if we did not already get the rank in parent of this
	       process, get it now. */
	    if (p_rank < 0)
		MPI_Comm_rank(hwdesc[hwlevel-1].comm, &p_rank);

	    /* Only the rank=0 process in the parent has the gather results.
	       Condense the information into just the ranks of the
	       process leaders */
	    if (p_rank == 0) {
		int j = 0;
		for (int i=0; i<p_size; i++) {
		    if (allleaders[i] >= 0) {
			if (j >= hwdesc[hwlevel].nDistinct) {
			    fprintf(stderr,
		      "Found too many leaders at index %d; expected only %d\n",
				    i, hwdesc[hwlevel].nDistinct);
			    fflush(stderr);
			}
			leaders[j++] = allleaders[i];
		    }
		}
	    }
	    DBG1("Bcast the leaders array of size %d\n",
		 hwdesc[hwlevel].nDistinct);
	    MPI_Bcast(leaders, hwdesc[hwlevel].nDistinct, MPI_INT,
		      0, hwdesc[hwlevel-1].comm);
	    hwdesc[hwlevel].leaders = leaders;
	    DBG0("Bcast leaders complete\n");

	    /* If there are as many comms as processes in the parent,
	     they all have to be of size one */
	    if (hwdesc[hwlevel].nDistinct == p_size)
		hwdesc[hwlevel].allsizeone = 1;
	    else
		hwdesc[hwlevel].allsizeone = 0;
#endif
	}

	DBG1("Split comm at level %d\n", hwlevel);
	MPI_Info_create(&hwinfo);
	MPI_Comm_split_type(hwdesc[hwlevel].comm,
			    MPI_COMM_TYPE_HW_UNGUIDED,
			    lrank,   /* Keep same process ordering as parent */
			    hwinfo,
			    &hwdesc[hwlevel+1].comm);

	buflen = 0;
	DBG0("Get resource string\n");
	{
	    char tmpstr[2];
	    /* MPICH requires a non-null location for the output string,
	       even when buflen is 0. So we provide tmpstr and ignore it */
	    MPI_Info_get_string(hwinfo, "mpi_hw_resource_type",
				&buflen, tmpstr, &flag);
	}
	if (flag && buflen > 0) {
	    /* Allocate buffer for value */
	    value = (char *)malloc(buflen);
	    if (!value) {
		fprintf(stderr, "Unable to allocate %d bytes for info value\n",
			buflen);
		MPI_Abort(MPI_COMM_WORLD, 1);
	    }
	    MPI_Info_get_string(hwinfo, "mpi_hw_resource_type",
				&buflen, value, &flag);
	    hwdesc[hwlevel+1].descstr = value;
	}
	MPI_Info_free(&hwinfo);

	hwlevel ++;

#if 0
	if (hwlevel >= hwdescLen) {
	    hwdescLen *= 2;
	    DBG1("Expand stack to depth %d\n", hwdescLen);
	    hwdesc = (hwdesc_t *)realloc(hwdesc,hwdescLen);
	    if (!hwdesc) {
		fprintf(stderr, "Unable to realloc hwdesc on %d\n", wrank);
		MPI_Abort(MPI_COMM_WORLD, 0);
	    }
	}
#else
	if (hwlevel >= maxdepth) {
	    return 1;
	}
#endif
    }
    *depth = hwlevel;
    return 0;
}
#endif /* MPI_VERSION < 4 */


/*@ BENV_HwdescPrintArray - Print the LOCAL elements of an hwdesc_t array

Input parameters:
+ fp - File pointer for output
. hw - hwdes_t array from BENV_HwdescGetLocal
. hwdepth - Number of elements in hw (depth from BENV_HwdescGetLocal)
- prefix - String to prefix each output line

Notes:
This is a local routine. Different processes should output to separate files.
If 'stdout' or 'stderr' is used as the file pointer, output may be lost or
interleaved. It is better to send output to a separate file for each process,
or ensure output is serialized to a single file (and due to buffering in
parallel runtime systems, this can be hard to accomplish for 'stdout' or
'stderr').
 @*/
int BENV_HwdescPrintArray(FILE *fp, hwdesc_t *hw, int hwdepth,
			  const char *prefix)
{
    for (int i=0; i<hwdepth; i++) {
	int lsize, lrank;
	if (hw[i].comm != MPI_COMM_NULL) {
	    MPI_Comm_rank(hw[i].comm, &lrank);
	    MPI_Comm_size(hw[i].comm, &lsize);
	}
	else {
	    lrank = -1;
	    lsize = 0;
	}
	if (lrank >= 0) {
	    fprintf(fp, "%s: idx=%d. comm(r=%d,size=%d): %s\n",
		    prefix, i, lrank, lsize, hw[i].descstr);
	}
	if (hw[i].nDistinct > 0) {
	    fprintf(fp, "%s: nleaders = %d, idx=%d, parent r=",
		    prefix, hw[i].idx, hw[i].nDistinct);
	    if (hw[i].leaders) {
		for (int j=0; j<hw[i].nDistinct-1; j++) {
		    fprintf(fp, "%d,", hw[i].leaders[j]);
		}
		fprintf(fp, "%d\n", hw[i].leaders[hw[i].nDistinct-1]);
	    }
	    else {
		fputc('\n', fp);
	    }
	}
    }
    fflush(fp);
    return 0;
}

/*@ BENV_HwdescPrintAll - Generate a printed description of the hwdest_t
  description of the system

Input Parameters:
+ fp - FILE pointer for output
. pcomm - Communicator for all processes. Call is collective over pcomm
. hw - Hwdesc_t from 'BENV_HwdescGetLocal' or similar
. hwdepth - Number of entries in 'hw'
- depth - Used to control indentation of output. Use '0' unless a different
 indentation is desired

Notes:
This is a collective call over 'pcomm'. However, only the process with
rank 0 in 'pcomm' performs output.

  @*/
int BENV_HwdescPrintAll(FILE *fp, MPI_Comm pcomm, hwdesc_t *hw, int hwdepth,
			int depth)
{
    int lrank, csize, wrank, token=0, prank;
    static MPI_Group wgroup=MPI_GROUP_NULL;
    static const char *sblock = 0;

    DBGC2("printHWDesc(%d): Enter hwdepth=%d\n", depth, hwdepth, 1);
    if (hwdepth <= 0) return 0;

    MPI_Comm_rank(hw[0].comm, &lrank);
    if (pcomm != MPI_COMM_NULL)
	MPI_Comm_rank(pcomm, &prank);
    else
	prank = -1;

    DBGC3("printHWDesc(%d): lrank=%d prank=%d\n", depth, lrank, prank, 1);

    /* Are we a leader at this level? */
    if (lrank == 0) {
	const char *rmsg=0;
	int   *hwranks=0;
	int   psize;

	if (hw[0].allsizeone) {
	    int range[3];
	    if (prank == 0) {
		MPI_Comm_size(pcomm, &psize);
		hwranks = (int *)malloc(psize * sizeof(int));
		if (!hwranks) return -1;
	    }
	    /* Compress the representation of a collection of communicators,
	       all of size one.*/
	    MPI_Comm_rank(MPI_COMM_WORLD, &wrank);
	    MPI_Gather(&wrank, 1, MPI_INT, hwranks, 1, MPI_INT, 0, pcomm);
	    if (prank == 0) {
		if (compressListToRange(psize, hwranks, range) == 0) {
		    rmsg = hwdescrange(depth, range, hw[0].descstr);
		}
		else {
		    rmsg = hwdesclist(depth, psize, hwranks, hw[0].descstr);
		}
		sblock = BENV_ConcatStringsFree(sblock, rmsg);
		free(hwranks);
	    }
	}
	else {
	    /* Wait for token *unless* we are the root process
	       OR send-to-self */
	    if (prank > 0) {
		DBGC1("printHWDesc(%d): wait for token\n", depth, 1);
		MPI_Recv(&token, 1, MPI_INT, MPI_ANY_SOURCE, 0, pcomm,
			 MPI_STATUS_IGNORE);
		DBGC1("printHWDesc(%d): token received\n", depth, 1);
	    }
	    /* output the description at this level */
	    MPI_Comm_size(hw[0].comm, &csize);
	    MPI_Comm_rank(MPI_COMM_WORLD, &wrank);
	    /* Format text describing this entry */
	    rmsg   = hwdescline(depth, csize, wrank, hw[0].idx, hw[0].descstr);
	    sblock = BENV_ConcatStringsFree(sblock, rmsg);
	}

	/* Depth first - descend into stack */
	if (hwdepth > 1) {
	    DBGC2("printHWDesc(%d): call self at depth %d\n",
		  depth, depth+1,1);
	    BENV_HwdescPrintAll(fp, hw[0].comm, hw+1, hwdepth-1, depth+1);
	}

	/* If we are the first of the leaders, invoke the other processes */
	if (!hw[0].leaders) {
	    DBGC1("printHWDesc(%d): Leaders is null\n", depth, 1);
	}
	else if (prank == hw[0].leaders[0]) {
	    if (hw[0].allsizeone) {
		;
		/* Nothing to do since all processes already contributed */
	    }
	    else {
		for (int i=1; i<hw[0].nDistinct; i++) {
		    /* Don't send to self */
		    /* Each of these is really an "invoke routine on the next
		       level for the leader */
		    DBGC2("printHWDesc(%d): call rank=%d\n",
			  depth, hw[0].leaders[i],1);
		    DBGC2("printHWDesc(%d): Call leader rank=%d (send)\n",
			  depth, hw[0].leaders[i], 1);
		    MPI_Send(&token, 1, MPI_INT, hw[0].leaders[i], 0, pcomm);
		    /* A more robust code would return a string, so that
		       only the master root process would write */
		    DBGC2("printHWDesc(%d): wait for return token from %d\n",
			  depth, hw[0].leaders[i], 1);
		    MPI_Recv(&token, 1, MPI_INT, hw[0].leaders[i], 0, pcomm,
			     MPI_STATUS_IGNORE);
		    DBGC2("printHWDesc(%d): Return from leader=%d\n",
			  depth, hw[0].leaders[i], 1);
		    rmsg = BENV_RecvString(hw[0].leaders[i], 0, pcomm);
		    sblock = BENV_ConcatStringsFree(sblock, rmsg);
		}
	    }
	}
	if (prank > 0 && !hw[0].allsizeone) {
	    DBGC1("printHWDesc(%d): return token\n", depth, 1);
	    MPI_Send(&token, 1, MPI_INT, 0, 0, pcomm);
	    BENV_SendString(sblock, 0, 0, pcomm);
	    free((void *)sblock);
	}
	else if (sblock && depth == 0) {
	    fputs(sblock, fp);
	    free((void *)sblock);
	}
    }
    else {
	/* Depth first - descend into stack */
	DBGC2("printHWDesc(%d): descend into stack at depth %d\n",
	      depth, depth+1, 1);
	BENV_HwdescPrintAll(fp, hw[0].comm, hw+1, hwdepth-1, depth+1);
    }

    DBGC1("printHWDesc(%d): return\n", depth, 1);
    return 0;
}

/* Given an array of integer values, try to compress it to a range
   or a strided range, and return that. Return 0 on success and 1 on failure */
static int compressListToRange(int n, int vec[], int range[3])
{
    int s, stride;

    if (n <= 1) return 1;
    s  = vec[0];
    stride = vec[1] - s;
    s = vec[1];
    for (int i=2; i<n; i++) {
	if (s + stride != vec[i]) return 1;
	s = vec[i];
    }
    range[0] = vec[0];
    range[1] = stride;
    range[2] = vec[n-1];

    return 0;
}

static const char *hwdescline(int indent, int csize, int wrank, int idx,
			      const char *descstr)
{
    int dlen = strlen(descstr), totlen, rc;
    char *str, *p;

    /* This length is an estimate -
       21 char + 2 ints + 15 char + int + descstr + null */
    totlen = 21 + 10 + 10 + 15 + 10 + dlen + 1;
    str = (char *)malloc(totlen*sizeof(char));
    if (!str) return 0;

    p = str;
    for (int i=0; i<indent; i++)
	*p++ = ' ';
    rc = snprintf(p, totlen-indent,
		  "comm(size=%d,wrank=%d,idx in parent=%d) %s\n",
		  csize, wrank, idx, descstr);
    if (rc < 0) return 0;

    return (const char *)str;
}

static const char *hwdescrange(int indent, int range[3], const char *descstr)
{
    int dlen = strlen(descstr), totlen, rc;
    char *str, *p;

    /* This length is an estimate - 10 char + 3 ints + descstr + null */
    totlen = 10 + 3*10 + dlen + 1;
    str = (char *)malloc(totlen*sizeof(char));
    if (!str) return 0;

    p = str;
    for (int i=0; i<indent; i++)
	*p++ = ' ';
    rc = snprintf(p, totlen-indent, "range: %d:%d:%d %s\n",
		  range[0], range[1], range[2], descstr);
    if (rc < 0) return 0;

    return (const char *)str;

}

static const char *hwdesclist(int indent, int n, int vec[], const char *descstr)
{
    int dlen = strlen(descstr), totlen, rc;
    char *str, *p;

    /* This length is an estimate -
       10 char + n ints and commas + descstr + null */
    totlen = 10 + n*(10+1) + dlen + 1;
    str = (char *)malloc(totlen*sizeof(char));
    if (!str) return 0;

    p = str;
    for (int i=0; i<indent; i++)
	*p++ = ' ';
    totlen -= indent;
    for (int i=0; i<n-1; i++) {
	rc = snprintf(p, totlen, "%d,", vec[i]);
	if (rc < 0) return 0;
	p += rc;
	totlen -= rc;
    }
    rc = snprintf(p, totlen, "%d %s\n", vec[n-1], descstr);
    if (rc < 0) return 0;

    return (const char *)str;
}

/* This is an internal routine, available to other parts of baseenv, which
   fills in a single hwdesc_t entry with information about the split of
   the parent communicator pcomm into disjoint comms, of which this process
   is an element of comm. The call is collective over pcomm.
   It sets the elemennts:
     nDistinct - number of distinct comm
     leaders   - Array of ranks in pcomm of the leaders (rank 0 in comm)
                 of the communicators comm (that resulted from a split of
		 pcomm)
     allsizeone - True if all comm are of size one.
*/
int BENVi_FindLeadersInSplit(MPI_Comm pcomm, MPI_Comm comm, hwdesc_t *hw)
{
    int *allleaders, *leaders, p_size, p_rank, is_lead, lrank, idx;

    MPI_Comm_rank(comm, &lrank);
    DBG0("find leaders\n");
    is_lead = lrank == 0;
    MPI_Allreduce(&is_lead, &hw->nDistinct, 1, MPI_INT,
		  MPI_SUM, pcomm);
    DBG1("Number of distinct comms is %d\n", hw->nDistinct);

    /* Gather up the rank in the parent of the leaders of the
       communicators at this level. Algorithm is a simple
       brute-force approach - gather ranks for *all* processes
       at this level (array size of the parent comm) and look for
       ranks >= 0 (use -1 if a process is not a leader on this
       level) */
    MPI_Comm_size(pcomm, &p_size);
    allleaders = (int *)malloc(p_size * sizeof(int));
    if (!allleaders) {
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
    if (is_lead)
	MPI_Comm_rank(pcomm, &p_rank);
    else
	p_rank = -1;
    DBG0("Gather ranks of leaders in parent\n");
    MPI_Gather(&p_rank, 1, MPI_INT, allleaders, 1, MPI_INT, 0, pcomm);

    /* On the root, compress this to remove the -1, then broadcast
       to the rest of the processes */
    DBG0("Bcast leaders\n");
    leaders = (int *)malloc(hw->nDistinct * sizeof(int));
    if (!leaders) {
	MPI_Abort(MPI_COMM_WORLD, 0);
    }
    /* See above - if we did not already get the rank in parent of this
       process, get it now. */
    if (p_rank < 0)
	MPI_Comm_rank(pcomm, &p_rank);

    /* Only the rank=0 process in the parent has the gather results.
       Condense the information into just the ranks of the
       process leaders */
    if (p_rank == 0) {
	int j = 0;
	for (int i=0; i<p_size; i++) {
	    if (allleaders[i] >= 0) {
		if (j >= hw->nDistinct) {
		    fprintf(stderr,
		    "Found too many leaders at index %d; expected only %d\n",
			    i, hw->nDistinct);
		    fflush(stderr);
		}
		leaders[j++] = allleaders[i];
	    }
	}
    }
    DBG1("Bcast the leaders array of size %d\n",
	 hw->nDistinct);
    MPI_Bcast(leaders, hw->nDistinct, MPI_INT, 0, pcomm);
    hw->leaders = leaders;
    DBG0("Bcast leaders complete\n");
    DBG0("Bcast my idx to others in my comm\n");
    if (lrank == 0) {
	/* Find my parent rank in leaders */
	for (idx=0; idx<hw->nDistinct; idx++)
	    if (leaders[idx] == p_rank) break;
	/* Sanity check */
	if (idx == hw->nDistinct) {
	    fprintf(stderr, "PANIC: Did not find leader\n");
	    MPI_Abort(comm, 1);
	    return 1;
	}
    }
    MPI_Bcast(&idx, 1, MPI_INT, 0, comm);
    hw->idx = idx;

    /* If there are as many comms as processes in the parent,
       they all have to be of size one */
    if (hw->nDistinct == p_size)
	hw->allsizeone = 1;
    else
	hw->allsizeone = 0;

    return 0;
}
