#include "smptestconf.h"
/*
 * Copyright (C) by University of Illinois 2023
 */
#include <stdlib.h>
#include <string.h>
#include "mpi.h"
#include "benvutil.h"

/*@ BENV_ConcatStringsFree - Concatenate two strings and return the result

Input Parameters:
+ s1 - First string. May be the null string. If not null, must be null
  terminated.
- s2 - Second string. May be the null string. If not null, must be null
  terminated.

Return Value:
 A null terminated string containting the concatenation of 's1' and 's2'.
 If 's1' and 's2' are both null, an empty string (a single null terminator)
 is returned. On error, a null is returned.

Notes:
's1' and 's2' are freed with 'free' if non-null.
@*/
const char *BENV_ConcatStringsFree(const char *s1, const char *s2)
{
    int l1=0, l2=0;
    char *str, *p;
    const char *pin;

    if (s1) l1 = strlen(s1);
    if (s2) l2 = strlen(s2);
    str = (char *)malloc(l1 + l2 + 1);
    if (!str) return 0;

    p   = str;
    if (s1) {
	pin = s1;
	while (*pin) *p++ = *pin++;
	free((void *)s1);
    }
    if (s2) {
	pin = s2;
	while (*pin) *p++ = *pin++;
	free((void *)s2);
    }
    *p = 0;
    return (const char *)str;
}

/*@ BENV_SendString - Send a string to another process using MPI Send

Input parmeters:
+ s - String to send. Must be null terminated
. dest - Rank of destination process
. tag  - Message tag
- comm - Communicator for message

Notes:
This routine sends a null-terminated string of arbitrary length to another
process. Blocking 'MPI_Send' is used.  If message isolation is needed,
the user should create a separate communicator.
@*/
int BENV_SendString(const char *s, int dest, int tag, MPI_Comm comm)
{
    int rc;
    rc = MPI_Send(s, strlen(s)+1, MPI_CHAR, tag, dest, comm);
    return rc;
}

/*@ BENV_RecvString - Receive a string from another process using MPI Recv

Input parmeters:
+ source - Rank of sending process
. tag  - Message tag
- comm - Communicator for message

Return Value:
A pointer to the string that was received.  The memory for this string is
allocated with 'malloc' and it is the responsibility of the user to free
the memory when it is no longer needed.  On error, a null string is returned.

Notes:
This routine receives a null-terminated string of arbitrary length
from another process. Blocking 'MPI_Recv' is used.  If message isolation
is needed, the user should create a separate communicator.
 @*/
const char *BENV_RecvString(int source, int tag, MPI_Comm comm)
{
    MPI_Status st;
    int ln, rc;
    char *str;

    rc = MPI_Probe(source, tag, comm, &st);
    rc = MPI_Get_count(&st, MPI_CHAR, &ln);
    str = (char *)malloc(ln);
    if (!str) return 0;
    rc = MPI_Recv(str, ln, MPI_CHAR, source, tag, comm, MPI_STATUS_IGNORE);
    return (const char *)str;
}
