#include "smptestconf.h"
/*
 * Copyright (C) by University of Illinois 2023
 */
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "mpi.h"
#define BENV_WITHMPI 1
#include "benvutil.h"

/* Used to debug node decomposition routines by overriding the hw topology
   inquiry routines to specify a decomposition */
static int cvar_node_ppn = -1;
static int cvar_node_spn = -1;
static int cvar_node_debug = 0;

/*@ BENV_NodeInitCvar - Initialize the control variables for the nodeinfo
    routines from environment variables
  @*/
int BENV_NodeCvarInit(void)
{
    char *envstr;
    envstr = getenv("BENV_NODEINFO_DEBUG");
    cvar_node_debug = envstr && (strcmp(envstr, "true") == 0);

    /* Processes per node */
    envstr = getenv("BENV_NODEINFO_DEBUG_PPN");
    if (envstr) {
	/* We define the processes per node, because that is more commonly
	   a property of a system, while the number of nodes is something
	   that the user implicitly selects by selecting the total number
	   of processes. */
	cvar_node_ppn = atoi(envstr);
	cvar_node_debug = 1;
    }
    /* Sockets per node */
    envstr = getenv("BENV_NODEINFO_DEBUG_SPN");
    if (envstr) {
	if (cvar_node_ppn > 0)
	    cvar_node_spn = atoi(envstr);
	else {
	    fprintf(stderr, "spn requires ppn be set\n"); fflush(stderr);
	}
    }

    return 0;
}

/*@ BENV_NodeCvarSet - Set debugging control variables for node info routines

Input Parameters:
+ name - Name of cvar (valid are ppn, debug, sockets)
- value - Value for cvar
  @*/
void BENV_NodeCvarSet(const char *name, int value)
{
    if (strcmp(name, "ppn") == 0)
	cvar_node_ppn     = value;
    else if (strcmp(name, "debug") == 0)
	cvar_node_debug   = value;
    else if (strcmp(name, "sockets") == 0)
	cvar_node_spn     = value;
    else {
	fprintf(stderr, "Unrecognized cvar %s\n", name); fflush(stderr);
    }
}

/*
 */
static int BENVi_NodeSplitDebug(MPI_Comm comm, hwdesc_t hw[], int maxdepth,
				int *depth);

/*@ BENV_NodeGetComms - Return communicators for nodes and xx

Input parameters:
+ comm - Communicator of processes for which to find the node and other
communicators
- maxdepth - size (number of elements) of the hw array
Output parameters:
+ hw - Array of information about the ...
- depth - Number of elements set in hw; that is, entries hw[0]..hw[depth-1]
have values

Special Features:
If the environment variable 'BENV_NODEINFO_DEBUG' is set to 'true', this routine
will ignore the hardware and return a decomposition of nodes based instead
on the value 'BENV_NODEINFO_DEBUG_PPN', which should be a string representing
a positive integer. This is used to debug applications that depend on a
decomposition into nodes without requiring that many physical nodes.

Notes:
This is a convenience function that provides node, socket(package) and
potentially additional information about the processes in a communicator.

Using this information, a new communicator can be constructed that maps
processes to nodes, sockets, and cores in any desired way, such as
contiguous ranks on sockets, then nodes.

The entries of hw are
.vb
    hw[0].comm - same as input comm
    hw[1].comm - communicator of processes on the same node
    hw[2].comm - communicator of processes on the same socket or other
                 hardware region
    hw[3].comm - communicator of processes on the same memory region
                 (potentially numaregion, e.g., cores sharing the same
                 L2 cache)
.ve
The specifics of the entries will depend on the MPI implementation and
the hardware environment

Implementation Notes:
This routine prefers to use BENV_HwdescGetLocal, which uses the
MPI_COMM_TYPE_HW_UNGUIDED split type to create the hw array. If that
is not available, or if it does not provide similarly useful data,
then MPI_COMM_TYPE_SHARED is used, along with attempts at determining
core and socket.

Not currently implemented is the use of hwloc to access information about
the hardware configuration.
  @*/
int BENV_NodeGetComms(MPI_Comm comm, hwdesc_t hw[], int maxdepth, int *depth)
{
    int rc;

    /* Debug set through BENV_NodeCvarInit or CvarSet */
    if (cvar_node_debug) {
	rc = BENVi_NodeSplitDebug(comm, hw, maxdepth, depth);
    }
    else {
	rc = BENV_HwdescGetLocal(comm, hw, maxdepth, depth);

	if (rc != 0 || *depth < 2) {
	    rc = BENV_HwdescGetByShared(comm, hw, maxdepth, depth);
	}
    }

    return 0;
}

int BENV_HwdescGetByShared(MPI_Comm comm, hwdesc_t hw[], int maxdepth,
			   int *depth)
{
    int hwlevel = 0;

    hw[0].comm    = comm;
    /* Try to get the communicator name as the description. */
    {
	char cname[MPI_MAX_OBJECT_NAME];
	int  cnameLen;
	MPI_Comm_get_name(comm, cname, &cnameLen);
	if (cnameLen > 0)
	    hw[0].descstr = strdup(cname);
	else
	    hw[0].descstr = strdup("Top level communicator");
    }
    hw[0].nDistinct  = 1;
    hw[0].leaders    = 0; /* No parent so no leaders array */
    hw[0].idx        = 0;
    hw[0].allsizeone = 0;

    /* Try to get nodes based on type_shared */
    MPI_Comm_split_type(comm, MPI_COMM_TYPE_SHARED, 0,
			MPI_INFO_NULL, &hw[1].comm);
    /* We use node here because this is a common result and this matches
       expectations for other codes */
    hw[1].descstr = strdup("node");

    BENVi_FindLeadersInSplit(comm, hw[1].comm, &hw[1]);
    *depth = 2;

    return 0;
}

/* Use for debugging - create a simple split based on environment variables */
static int BENVi_NodeSplitDebug(MPI_Comm comm, hwdesc_t hw[], int maxdepth,
				int *depth)
{
    int hwlevel = 0, wsize, wrank, nodes, node, pinnode;

    hw[0].comm    = comm;
    /* Try to get the communicator name as the description. */
    {
	char cname[MPI_MAX_OBJECT_NAME];
	int  cnameLen;
	MPI_Comm_get_name(comm, cname, &cnameLen);
	if (cnameLen > 0)
	    hw[0].descstr = strdup(cname);
	else
	    hw[0].descstr = strdup("Top level communicator");
    }
    hw[0].nDistinct  = 1;
    hw[0].leaders    = 0; /* No parent so no leaders array */
    hw[0].idx        = 0;
    hw[0].allsizeone = 0;

    /* Get ppn (processes per node) and split into nodes. */
    MPI_Comm_size(comm, &wsize);
    MPI_Comm_rank(comm, &wrank);

    if (cvar_node_ppn > 0) {
	nodes = wsize / cvar_node_ppn;
    }
    else {
	/* Default is 2 nodes */
	nodes = 2;
	cvar_node_ppn = wsize / 2;
    }
    /* Given the  number of nodes and processes per node (ppn), determine
       the node and process in node (pinnode) of this process, based on
       its rank */
    node    = wrank % nodes;
    pinnode = wrank % cvar_node_ppn;

    /* Try to get nodes based on type_shared */
    MPI_Comm_split(comm, node, pinnode, &hw[1].comm);
    /* We use node here because this is a common result and this matches
       expectations for other codes */
    hw[1].descstr = strdup("node");

    BENVi_FindLeadersInSplit(comm, hw[1].comm, &hw[1]);
    *depth = 2;

    /* FIXME: Add optional socket/package level, and individual cores to
       match MPICH output */

    return 0;
}
